"""resconfig Configuration Library
==================================

:copyright: (c) 2020 by Taro Sato.
:license: Apache 2.0, see LICENSE.txt for more details.
"""

from .actions import Action  # noqa
from .resconfig import ResConfig  # noqa

__author__ = "Taro Sato"
__author_email__ = "okomestudio@gmail.com"
__version__ = "20.4.2a"
__license__ = "Apache-2.0"
